/**************************************************
HXDbstjudge: bst judgement for the light curve fits 

ver 0.0.1 K.Yamaoka  2006 Feb. 12 
   first version to make fits light curve
ver 0.0.2 K.Yamaoka  2009 May 6 
   bug fix in hxdgrb machine
ver 0.0.3 K.Yamaoka  2009 May 7 
   further refinement in bst judgment
ver 0.0.4 K.Yamaoka  2009 May 9
   Disable for judgment during the WAM scan 

ver 0.0.5 C.Padgett  2009 Oct 6
   * Fixed std deviation calculation for disabling
     during WAM scan:
       - made it a two pass calculation, since NaN's 
         were happening with the one-pass version
   * Added burst duration calculation, and eliminated
     numerous bursts being reported for each burst
       - each burst is now integrated until no longer
         detected.
   * Added clobber parameter
   * Added comment header to output file:
       comment lines start with '#'
   * Use dynamic memory for reading + storing
     light curve
   * Added TIMEZERO and TIMEPIXR to time calculations
   * Included FRACEXP in timing/rate calculations
   * In hxdbstjudge.def, included HXDleapsecInit to
     handle user supplied (or HEASoft/CALDB) leap
     second file. Also added leapfile parameter.
   * Renamed module HXDbstJudge
   * Renamed integ_time parameter to bgd_integ_time
   * Added lots of comments
   
***************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <cli.h>
#include <com.h>
#include <bnk.h>
#include <evs.h>
#include <anl.h>
#include <atFunctions.h>
#include <aste_time.h>
#include <fitsio.h>
#include <cfortran.h>
#include <hbook.h>
#include "hxd/HXD.h"
#include "pil.h"

#define DEBUG 0

/*
** task name and version
*/
static char pname[ ]        = "HXDbstJudge";
char HXDbstJudge_version[ ] = "version 0.0.5";

/*
** input parameters
*/
static char filename[ FLEN_FILENAME ];
static char newfilename[ FLEN_FILENAME ];

static int clobber        = 0;
static int bgd_integ_time = 0;

static double sigma        = 0.0;
static double judge_time   = 0.0;

/*
** output text file pointer
*/
FILE *text_fp = ( FILE* ) 0;

/*
** light curve vars
*/
static long num_row = 0;

static double timedel      = 0.0;
static double timezero     = 0.0;
static double timepixr     = 0.0;

static double* lctime  = ( double* )0;
static double* rate    = ( double* )0; 
static double* error   = ( double* )0;
static double* fracexp = ( double* )0; 

/*
** columns to read from light curve
*/
#define LCFITS_NUM_COL 4
enum {
    TIME = 0,
    RATE,
    ERROR,
    FRACEXP
};
static char* lcfits_colname[ LCFITS_NUM_COL ] = {
    "TIME",
    "RATE",
    "ERROR",
    "FRACEXP"
};

/*
** lcfits_read
**
**      reads a fits light curve
**      must have TIME, RATE, ERROR and FRACEXP columns
*/
static int
lcfits_read( char filename[ ] )
{
    char comment[ 80 ];

    char* k;
    
    int i;
    int istat     = 0;
    int hdutype   = 0;
    int anynul    = 0;
    int caseinsen = TRUE;
    
    int colnum[ LCFITS_NUM_COL ];

    long firstlow  = 1;
    long firstelem = 1;

    double nulval = 0.0;

    fitsfile* fp = ( fitsfile* )0;

    anl_msg_info( "%s: reading light curve in %s\n",
                  pname, filename );
    /*
    ** open and go to 1st extension (assuming it's a light curve!)
    */
    if ( fits_open_file( &fp, filename, READONLY, &istat ) ) {
        anl_msg_error( "%s: fits_open_file('%s') failed (%d)\n",
                       pname, filename, istat );
        return ANL_NG;
    } else {
        if ( fits_movabs_hdu( fp, 2, &hdutype, &istat ) ) {
            anl_msg_error( "%s: fits_movabs_hdu(2) failed on %s (%d)\n",
                           pname, filename, istat );
            return ANL_NG;
        }
    }

    /*
    ** get column numbers
    */
    for( i = 0; i < LCFITS_NUM_COL; i++ ) {
        if ( fits_get_colnum( fp, caseinsen, lcfits_colname[ i ],
                              &colnum[ i ], &istat ) ) {
            anl_msg_error( "%s: fits_get_colnum('%s') failed on %s (%d)\n",
                           pname, lcfits_colname[ i ], filename, istat );
            return ANL_NG;
        }
    }

    /*
    ** read some keys
    */
    if ( fits_read_key_dbl( fp, k = "TIMEDEL",  &timedel, comment, &istat ) ) {
        anl_msg_error( "%s: fits_read_key_dbl('%s') failed on %s (%d)\n",
                       pname, k, filename, istat );
        return ANL_NG;
    }
    if ( fits_read_key_dbl( fp, k = "TIMEZERO",  &timezero, comment, &istat ) ) {
        istat = 0;
        timezero = 0.0;
        fits_read_errmsg( comment );
    }
    if ( fits_read_key_dbl( fp, k = "TIMEPIXR",  &timepixr, comment, &istat ) ) {
        istat = 0;
        timepixr = 0.5;
        fits_read_errmsg( comment );
    }
    if ( fits_read_key_lng( fp, k = "NAXIS2",  &num_row, comment, &istat) ) {
        anl_msg_error( "%s: fits_read_key_dbl('%s') failed on %s (%d)\n",
                       pname, k, filename, istat );
        return ANL_NG;
    }

    /*
    ** allocate memory
    */
    lctime  = ( double* )calloc( num_row, sizeof( double ) );
    rate    = ( double* )calloc( num_row, sizeof( double ) );
    error   = ( double* )calloc( num_row, sizeof( double ) );
    fracexp = ( double* )calloc( num_row, sizeof( double ) );
    if ( lctime == NULL || rate == NULL || error == NULL || fracexp == NULL ) {
        anl_msg_error( "%s: failed to allocate memory for light curve\n",
                       pname );
        if ( lctime )  { free( lctime ); }
        if ( rate )    { free( rate ); }
        if ( error )   { free( error ); }
        if ( fracexp ) { free( fracexp ); }
        return ANL_NG;
    }

    /*
    ** read TIME, RATE, ERROR, and FRACEXP columns
    */
    if ( fits_read_col_dbl( fp, colnum[ TIME ], firstlow, firstelem, num_row,
                            nulval, lctime, &anynul, &istat ) ) {
        anl_msg_error( "%s: fits_read_col_dbl('%s') failed on %s (%d)\n",
                       pname, lcfits_colname[ TIME ], filename, istat );
        return ANL_NG; 
    }
    if ( fits_read_col_dbl( fp, colnum[ RATE ], firstlow, firstelem, num_row,
                            nulval, rate, &anynul, &istat ) ) {
        anl_msg_error( "%s: fits_read_col_dbl('%s') failed on %s (%d)\n",
                       pname, lcfits_colname[ RATE ], filename, istat );
        return ANL_NG; 
    }
    if ( fits_read_col_dbl( fp, colnum[ ERROR ], firstlow, firstelem, num_row,
                            nulval, error, &anynul, &istat ) ) {
        anl_msg_error( "%s: fits_read_col_dbl('%s') failed on %s (%d)\n",
                       pname, lcfits_colname[ ERROR ], filename, istat );
        return ANL_NG; 
    }
    if ( fits_read_col_dbl( fp, colnum[ FRACEXP ], firstlow, firstelem, num_row,
                            nulval, fracexp, &anynul, &istat ) ) {
        anl_msg_error( "%s: fits_read_col_dbl('%s') failed on %s (%d)\n",
                       pname, lcfits_colname[ FRACEXP ], filename, istat );
        return ANL_NG; 
    }

    /*
    ** done
    */
    if ( fits_close_file( fp, &istat ) ) {
        anl_msg_error( "%s: fits_close_file('%s') failed (%d)\n",
                       pname, filename, istat );
        return ANL_NG;
    }
    return ANL_OK;
}

/*
** bst_judge -
**
**      looks for and categorizes bursts
*/
void
bst_judge( int num_row, int bgd_integ_time, double judge_time, 
           double sigma, int *bst_flag )
{

    int i, j, k;

    int judge_dis_flag = 0;
    int burst_stop_ind = 0;
    
    static int nonzero_flag   = 0; 
    static int judge_dis_time = 0;
  
    double dd1, dd2;
    double total_bgcnt        = 0.0;
    double total_time         = 0.0;
    double total_bgcnt_square = 0.0;
    double total_ave_bg       = 0.0;
    double variance           = 0.0;
    double sum_bg             = 0.0;
    double cnt_src            = 0.0;
    double average            = 0.0;
    double poisson            = 0.0;
    double cnt_level          = 0.0;
    double sigma_level        = 0.0;
    double src_rate           = 0.0;
    double bgd_tot_time       = 0.0;
    double src_tot_time       = 0.0;
    double burst_start        = 0.0;
    double burst_stop         = 0.0;
    double burst_dur          = 0.0;
    double burst_cnts         = 0.0;
    double burst_tot_time     = 0.0;
    double burst_sigma        = 0.0;

    static double pre_sigma_level  = 0.0;
    static double pre_cnt_level    = 0.0;
    static double pre_cnt_src      = 0.0;
    static double pre_cnt_src_rate = 0.0;
    
    AtTimeD burstStart;
    AtTimeD burstStop;

    /*
    ** calculate the average rate and variance (stddev really)
    ** of all times where RATE > 1.0
    */
    anl_msg_info( "%s: calculating basic light curve statistics\n",
                  pname );
    for ( i = 0; i < num_row; i++ ) {
        if ( rate[ i ] > 1.0 ) {
            dd1 = timedel * fracexp[ i ];
            dd2 = rate[ i ] * dd1;
            total_bgcnt += dd2;
            total_bgcnt_square += dd2 * dd2;
            total_time += dd1;
        }
    }
    total_ave_bg = total_bgcnt / total_time;

    dd1 = 0.0;
    for ( i = 0; i < num_row; i++ ) {
        if ( rate[ i ] > 1.0 ) {
            variance += ( rate[ i ] - total_ave_bg ) * ( rate[ i ] - total_ave_bg );
            dd1 += 1.0;
        }
    }
    if ( dd1 > 0.0 ) {
        variance /= dd1;
    } else {
        variance = 0.0;
    }
    variance = sqrt( variance );
    anl_msg_info( "%s: mean count rate (above 1.0 c/s): %.4f +- %.4f c/s\n",
                  pname, total_ave_bg, variance );

    /*
    ** find bursts
    */
    for ( i = ( int )( bgd_integ_time / timedel ); i < num_row; i++ ) {

        judge_dis_flag = 0;

        /*
        ** calculate avarage for past "integ_time"
        ** only if we either don't have a burst or
        ** we have gone far enough past the last one
        */
        if ( burst_stop == 0.0 ||
             i >= burst_stop_ind + ( int )( bgd_integ_time / timedel ) ) {
            nonzero_flag = 1;
            sum_bg       = 0.0;
            bgd_tot_time = 0.0;
            for( j = i - ( int )( bgd_integ_time / timedel ); j < i; j++ ) {
                if ( rate[ j ] <= 1.0 ) {
                    nonzero_flag = 0;
                }
                if ( j >= 1 ) {
                    if ( rate[ j - 1 ] <= 1.0 ) {
                        nonzero_flag = 0;
                    }
                }
                sum_bg += rate[ j ] * timedel * fracexp[ j ];
                bgd_tot_time += timedel * fracexp[ j ];
            }

            /*
            ** average BG rate
            */
            if ( bgd_tot_time != 0.0 ) {
                average = sum_bg / bgd_tot_time;
            } else {
                average = 0.0;
            }
        }

        /*
        ** calculation for source times
        */
        cnt_src = 0.0;
        src_tot_time = 0.0;
        for ( k = i; k < i + ( int )( judge_time / timedel ); k++ ) {
            cnt_src += rate[ k ] * timedel * fracexp[ k ];
            src_tot_time += timedel * fracexp[ k ];
        }
        if ( src_tot_time != 0.0 ) {
            src_rate = cnt_src / src_tot_time;
        } else {
            src_rate = 0.0;
        }
        poisson     = sqrt( average * src_tot_time );
        cnt_level   = cnt_src - average * src_tot_time;
        sigma_level = cnt_level / poisson;

        /*
        ** probably during the WAM scan
        ** disable for 10 minutes
        */
        if ( pre_sigma_level < -10.0 &&
             src_rate < ( total_ave_bg - variance ) &&
             src_rate > 1.0 &&
             pre_cnt_src_rate < ( total_ave_bg - variance ) &&
             pre_cnt_src_rate > 1.0 ) {
            judge_dis_time = i + ( int )( 600.0 + bgd_integ_time ) / timedel;
        }

        /*
        ** update WAM scan check vars
        */
        pre_cnt_src     = cnt_src;
        pre_cnt_level   = cnt_level;
        pre_sigma_level = sigma_level;
        if ( src_tot_time != 0.0 ) {
            pre_cnt_src_rate = pre_cnt_src / src_tot_time;
        } else {
            pre_cnt_src_rate = 0.0;
        }

        /*
        ** check if there's a burst
        */
        if ( sigma_level >= sigma && nonzero_flag && i > judge_dis_time ) {

            *bst_flag = 1;

            /*
            ** start of burst
            */
            burst_start    = lctime[ i ] - timedel * timepixr + timezero;
            burst_cnts     = cnt_src;
            burst_tot_time = src_tot_time;

            /*
            ** loop until our burst condition is no longer met
            ** and we're not in the WAM scan
            ** Calculate the running rate in the judgetime window
            */
            j = i;
            while ( pre_sigma_level > -10.0 &&
                    src_rate >= ( total_ave_bg - variance ) &&
                    src_rate > 1.0 &&
                    pre_cnt_src_rate >= ( total_ave_bg - variance ) &&
                    pre_cnt_src_rate > 1.0 &&
                    sigma_level >= sigma &&
                    k < num_row ) {

                /*
                ** accumulate total burst counts and time
                */
                burst_cnts     += rate[ k ] * timedel * fracexp[ k ];
                burst_tot_time += timedel * fracexp[ k ];

                /*
                ** add the next point to the running total, subtract the oldest
                */
                cnt_src += rate[ k ] * timedel * fracexp[ k ];
                cnt_src -= rate[ i ] * timedel * fracexp[ i ];

                src_tot_time += timedel * fracexp[ k ];
                src_tot_time -= timedel * fracexp[ i ];

                if ( src_tot_time != 0.0 ) {
                    src_rate = cnt_src / src_tot_time;
                } else {
                    src_rate = 0.0;
                }

                /*
                ** get new sigma
                */
                poisson     = sqrt( average * src_tot_time );
                cnt_level   = cnt_src - average * src_tot_time;
                sigma_level = cnt_level / poisson;

                /*
                ** update WAM scan check vars
                */
                pre_cnt_src     = cnt_src;
                pre_cnt_level   = cnt_level;
                pre_sigma_level = sigma_level;
                if ( src_tot_time != 0.0 ) {
                    pre_cnt_src_rate = pre_cnt_src / src_tot_time;
                } else {
                    pre_cnt_src_rate = 0.0;
                }

                k++;
                i++;
            }

            /*
            ** warn if we didn't add any bins
            */
            if ( k == j + 1 + ( int )( judge_time / timedel ) ) {
                anl_msg_info( "%s: WARNING: burst duration may be over-estimated at %.16E\n",
                              pname, burst_start );
            }

            /*
            ** subtract the last addition (the one that made us exit the loop)
            ** from the total burst measures
            */
            burst_cnts     -= rate[ k - 1 ] * timedel * fracexp[ k - 1 ];
            burst_tot_time -= timedel * fracexp[ k - 1 ];

            /*
            ** burst stop time, duration, total sigma
            */
            burst_stop_ind = k - 2;
            burst_stop     = lctime[ k - 2 ] + timedel * timepixr + timezero;
            burst_dur      = burst_stop - burst_start;
            poisson        = sqrt( average * burst_tot_time );
            burst_sigma    = ( burst_cnts - average * burst_tot_time ) / poisson;
            anl_msg_info( "%s: burst found from %.16E-%.16E (%.1f-sigma)\n",
                          pname, burst_start, burst_stop, burst_sigma );

            /*
            ** is the total burst a detection?
            */
            if ( burst_sigma >= sigma ) {
                
                /*
                ** calculate the attime
                */
                aste2attimeD( burst_start, &burstStart );
                aste2attimeD( burst_stop, &burstStop );

                /*
                ** write a line to the output file
                */
                fprintf( text_fp, "%04d-%02d-%02d %02d:%02d:%02d %.16E %.16E %.8E %8d %7.1f\n",
                         burstStart.yr, burstStart.mo, burstStart.dy, burstStart.hr, 
                         burstStart.mn, burstStart.sc, burst_start,
                         burst_stop, burst_dur, ( int )burst_cnts, burst_sigma );
            }

            i = k - 1;

        } else {
            *bst_flag = 0;
        }
    }
    return;
}

/*
** dummy
*/
void
HXDbstJudge_startup( int *status )
{
    *status = ANL_OK;
}

/*
** common startup - read parameters
*/
void
HXDbstJudge_com( int *status )
{

    if ( ( *status = PILGetFname( "input_name", filename ) ) ) {
        anl_msg_error( "%s: PILGetFname input_name error (%d)\n",
                       pname, *status );
        *status = ANL_ERROR;
        return;
    }
    if ( ( *status = PILGetString( "outfile", newfilename ) ) ) {
        anl_msg_error( "%s: PILGetString outfile error (%d)\n",
                       pname, *status );
        *status = ANL_ERROR;
        return;
    }
    if ( ( *status = PILGetInt( "bgd_integ_time", &bgd_integ_time ) ) ) {
        anl_msg_error( "%s: PILGetInt integ_time(%d)\n",
                       pname, *status );
        *status = ANL_ERROR;
        return;
    }
    if ( ( *status = PILGetReal( "delta_t", &judge_time ) ) ) {
        anl_msg_error( "%s: PILGetReal delta_t error (%d)\n",
                       pname, *status );
        *status = ANL_ERROR;
        return;
    }
    if ( ( *status = PILGetReal( "sigma", &sigma ) ) ) {
        anl_msg_error( "%s: PILGetReal sigma error (%d)\n",
                       pname, *status );
        *status = ANL_ERROR;
        return;
    }
    if ( ( *status = PILGetBool( "clobber", &clobber ) ) ) {
        anl_msg_error( "%s: PILGetBool clobber error (%d)\n",
                       pname, *status );
        *status = ANL_ERROR;
        return;
    }
    *status = ANL_OK;
}    

/*
** dummies
*/
void
HXDbstJudge_init( int *status )
{
    *status = ANL_OK;
}

void
HXDbstJudge_his( int *status )
{
    *status = ANL_OK;
}

/*
** begin run - reads light curve and calculates "background"
*/
void
HXDbstJudge_bgnrun( int *status )
{
    int istat  = 0;
    int exists = 0;

    /*
    ** check if we need to clobber
    */
    if ( fits_file_exists( newfilename, &exists, &istat ) ) {
        anl_msg_error( "%s: could not determine if %s exists!\n",
                       pname, newfilename );
        *status = ANL_ERROR;
        return;
    }
    if ( exists == 1 && !clobber ) {
        anl_msg_error( "%s: %s exists and clobber not set!\n",
                       pname, newfilename );
        *status = ANL_ERROR;
        return;
    }

    /*
    ** read the light curve
    */
    if ( lcfits_read( filename ) ) {
        anl_msg_error( "%s: failed to read light curve from %s\n",
                       pname, filename );
        *status = ANL_ERROR;
        return;
    }

    /*
    ** open output file
    */
    text_fp = fopen( newfilename, "w" );
    if ( text_fp == NULL ) {
        anl_msg_error( "%s: failed to open file %s\n",
                       pname, newfilename );
        *status = ANL_ERROR;
        return;
    }

    /*
    ** print header into output file
    */
    fprintf( text_fp, "\
#\n\
# %s %s\n\
# Detected bursts in %s\n\
#\n\
#   START DATE         TSTART (AETIME)        TSTOP (AETIME)       DURATION (s)    COUNTS   SIGMA\n\
#------------------------------------------------------------------------------------------------\n",
             pname, HXDbstJudge_version, filename );

    *status = ANL_OK;       
}

/* 
** analysis loop
*/
void
HXDbstJudge_ana( int nevent, int eventid, int *status )
{
    int bst_flag;

    /*
    ** detect bursts
    */
    bst_judge( num_row, bgd_integ_time, judge_time, sigma, &bst_flag );

    *status = ANL_QUIT;
}

/*
** cleanup
*/
void
HXDbstJudge_endrun( int *status )
{    

    /*
    ** close output file
    */
    if ( text_fp ) {
        fclose( text_fp );
    }

    /*
    ** free any allocated memory
    */
    if ( lctime )    { free( lctime ); }
    if ( rate )    { free( rate ); }
    if ( error )   { free( error ); }
    if ( fracexp ) { free( fracexp ); }

    *status = ANL_OK;
}

/*
** dummy
*/
void
HXDbstJudge_exit( int *status )
{
    *status = ANL_OK;
}
